/*-
 * Copyright (c) 2007 Akihiko GOTANDA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

/*
 * Driver for SD card controller
 * based on SD Host Controller Standard defined by SD Assiciation
 */

#include <dev/sdmmc/sdvar.h>
#include <dev/sdmmc/sdshcreg.h>

#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/malloc.h>
#include <machine/bus.h>

#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>

#include "sdbus_if.h"

int sdshc_timeout_interval = 0;
#define	DEFAULT_TIMEOUT_INTERVAL	(hz / 10)
int sdshc_polling_interval = 0;
#define	DEFAULT_POLLING_INTERVAL	(hz / 100)
int sdshc_wait_after_insertion = 0;
#define	DEFAULT_WAIT_AFTER_INSERTION	(hz)
int sdshc_disable_ti_alt_if = 1;
int sdshc_use_highspeed_mode = 1;
SYSCTL_NODE(_hw, OID_AUTO, sdshc, CTLFLAG_RD, 0, "SD card slot controller");
SYSCTL_INT(_hw_sdshc, OID_AUTO, timeout_interval, CTLFLAG_RW,
	   &sdshc_timeout_interval, 0, "interrupt timeout interval");
SYSCTL_INT(_hw_sdshc, OID_AUTO, polling_interval, CTLFLAG_RW,
	   &sdshc_polling_interval, 0, "status polling interval");
SYSCTL_INT(_hw_sdshc, OID_AUTO, wait_after_insertion, CTLFLAG_RW,
	   &sdshc_wait_after_insertion, 0,
	   "wait interval after card insertion");
SYSCTL_INT(_hw_sdshc, OID_AUTO, disable_ti_alt_if, CTLFLAG_RW,
	   &sdshc_disable_ti_alt_if, 0, "disable TI's UltraMedia controller");
SYSCTL_INT(_hw_sdshc, OID_AUTO, use_highspeed_mode, CTLFLAG_RW,
	   &sdshc_use_highspeed_mode, 0, "use high speed mode if usable");

#if defined(SD_DEBUG)
int sdshc_debug_level = 0x10;
SYSCTL_INT(_debug, OID_AUTO, sdshc, CTLFLAG_RW,
	   &sdshc_debug_level, 0, "sdshc debug level");
#endif

struct sdshc_ident {
	uint16_t	vendor;
	uint16_t	device;
	const char	*name;
	uint32_t	quirk_device;
};


/*
 * TI's UltraMedia controller
 */
#if defined(USE_TI_DUMMY_DRIVER)
static const struct sdshc_ident sdshc_ti_alt_ident_table[] = {
    { 0x104c, 0x8033, "Texas Instruments PCIxx21 Integrated FlashMedia Controller", 0 },	/* Not yet tested */
    { 0x104c, 0x803b, "Texas Instruments 5-in-1 Multimedia Card Reader (SD/MMC/MS/MS PRO/xD)", 0 },	/* Not yet tested */
    { 0, 0, NULL }
};

static int	sdshc_ti_alt_probe(device_t);
static int	sdshc_ti_alt_attach(device_t);
static int	sdshc_ti_alt_detach(device_t);

static device_method_t sdshc_ti_alt_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		sdshc_ti_alt_probe),
	DEVMETHOD(device_attach,	sdshc_ti_alt_attach),
	DEVMETHOD(device_detach,	sdshc_ti_alt_detach),

	{ 0, 0 }
};

static driver_t sdshc_ti_alt_driver = {
	"sdshc_dummy",
	sdshc_ti_alt_methods,
	0
};

static devclass_t sdshc_ti_alt_devclass;

DRIVER_MODULE(sdshc_dummy, sdshc, sdshc_ti_alt_driver, sdshc_ti_alt_devclass, NULL, NULL);


static int
sdshc_ti_alt_probe(device_t dev)
{
	const struct sdshc_ident *ident;

	for (ident = sdshc_ti_alt_ident_table; ident->name != NULL; ident++) {
		if (pci_get_vendor(dev) == ident->vendor &&
		    pci_get_device(dev) == ident->device) {
			device_set_desc(dev, ident->name);
			return 0;
		}
	}

	return ENXIO;
}


static int
sdshc_ti_alt_attach(device_t dev)
{
    uint32_t reg;

    MSG((dev, "disabling TI's alternate FlushMedia controller\n"));
    reg = pci_read_config(dev, 0x4c, 1);
    reg |= 0x02;
    pci_write_config(dev, 0x4c, reg, 1);

    return 0;
}


static int
sdshc_ti_alt_detach(device_t dev)
{
    /* do nothing */

    return 0;
}

#endif	/* defined(USE_TI_DUMMY_DRIVER) */


static const struct sdshc_ident sdshc_ident_table[] = {
    { 0x1180, 0x0822, "Ricoh Co Ltd R5C822 SD/SDIO/MMC/MS/MSPro Host Adapter", 0x00000000 },
    { 0x104c, 0x8034, "Texas Instruments PCIxx11/21 SDA Standard Compliant SD Host Controller", 0x104c8033 },	/* Not yet tested */
    { 0x104c, 0x803c, "Texas Instruments PCIxx12 SDA Standard Compliant SD Host Controller", 0x104c803b },	/* Not yet tested */
    { 0, 0, NULL }
};


static int	sdshc_probe(device_t);
static int	sdshc_attach(device_t);
static int	sdshc_detach(device_t);
static int	sdshc_shutdown(device_t);
static int	sdshc_suspend(device_t);
static int	sdshc_resume(device_t);

static void	sdshc_intr(void *);


static device_method_t sdshc_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		sdshc_probe),
	DEVMETHOD(device_attach,	sdshc_attach),
	DEVMETHOD(device_detach,	sdshc_detach),
	DEVMETHOD(device_shutdown,	sdshc_shutdown),
	DEVMETHOD(device_suspend,	sdshc_suspend),
	DEVMETHOD(device_resume,	sdshc_resume),

	{ 0, 0 }
};

static driver_t sdshc_driver = {
	"sdshc",
	sdshc_methods,
	sizeof(struct sdshc_softc)
};

static devclass_t sdshc_devclass;
#define SDSHC_SOFTC(unit) \
	((struct sdshc_softc*)devclass_get_softc(sdshc_devclass, unit))

DRIVER_MODULE(sdshc, pci, sdshc_driver, sdshc_devclass, NULL, NULL);


#define	SDSHC_REG_READ1(sc, ofs)	\
	bus_space_read_1((sc)->bt, (sc)->bh, ofs)
#define	SDSHC_REG_READ2(sc, ofs)	\
	bus_space_read_2((sc)->bt, (sc)->bh, ofs)
#define	SDSHC_REG_READ4(sc, ofs)	\
	bus_space_read_4((sc)->bt, (sc)->bh, ofs)
#define	SDSHC_READ1(sc, name)		SDSHC_REG_READ1(sc, SDSHC_HSR_ ## name)
#define	SDSHC_READ2(sc, name)		SDSHC_REG_READ2(sc, SDSHC_HSR_ ## name)
#define	SDSHC_READ4(sc, name)		SDSHC_REG_READ4(sc, SDSHC_HSR_ ## name)

#define	SDSHC_REG_WRITE1(sc, ofs, val)	\
	bus_space_write_1((sc)->bt, (sc)->bh, ofs, val)
#define	SDSHC_REG_WRITE2(sc, ofs, val)	\
	bus_space_write_2((sc)->bt, (sc)->bh, ofs, val)
#define	SDSHC_REG_WRITE4(sc, ofs, val)	\
	bus_space_write_4((sc)->bt, (sc)->bh, ofs, val)
#define	SDSHC_WRITE1(sc, name, val)	SDSHC_REG_WRITE1(sc, SDSHC_HSR_ ## name, val)
#define	SDSHC_WRITE2(sc, name, val)	SDSHC_REG_WRITE2(sc, SDSHC_HSR_ ## name, val)
#define	SDSHC_WRITE4(sc, name, val)	SDSHC_REG_WRITE4(sc, SDSHC_HSR_ ## name, val)

static int	sdshc_init(void *);
static int	sdshc_config(struct sdshc_softc *sc);
static int	sdshc_reset(struct sdshc_softc *sc);
static void 	sdshc_stop(struct sdshc_softc *sc);


static d_open_t  sdshc_open;
static d_close_t sdshc_close;
/*
static d_poll_t  sdshc_poll;
static d_purge_t sdshc_purge;
*/

static struct cdevsw sdshc_cdevsw = {
    .d_version = D_VERSION,
    .d_flags =	 D_NEEDGIANT,
    .d_open =	 sdshc_open,
    .d_close =	 sdshc_close,
/*
    .d_poll =	 sdshc_poll,
    .d_purge =	 sdshc_purge,
*/
    .d_name =	 "sdshc",
};


#define	SDSHC_UNIT(dev)	minor(dev)

static int
sdshc_probe(device_t dev)
{
	const struct sdshc_ident *ident;
	struct sdshc_softc *sc = device_get_softc(dev);

	for (ident = sdshc_ident_table; ident->name != NULL; ident++) {
		if (pci_get_vendor(dev) == ident->vendor &&
		    pci_get_device(dev) == ident->device) {
			device_set_desc(dev, ident->name);
			sc->quirk_device = ident->quirk_device;
			return 0;
		}
	}

	if (pci_get_class(dev) == PCIC_BASEPERIPH &&
	    pci_get_subclass(dev) == 0x05) {
	    switch (pci_get_progif(dev)) {
	      case 0x00:
		device_set_desc(dev, "Generic SD Host Controller w/o DMA");
		return 0;
	      case 0x01:
		device_set_desc(dev, "Generic SD Host Controller with DMA");
		return 0;
	    }
	}
	return ENXIO;
}


static int
sdshc_attach(device_t dev)
{
    struct sdshc_softc *sc = device_get_softc(dev);
    int cnt, error;
			
    sc->dev = dev;

    mtx_init(&sc->mtx, device_get_nameunit(dev), NULL,
	     MTX_DEF | MTX_RECURSE);
    sc->locked = 0;

    if (pci_get_powerstate(dev) != PCI_POWERSTATE_D0) {
	MSG((dev, "chip is in D%d power mode "
	     "-- setting to D0\n", pci_get_powerstate(dev)));
	pci_set_powerstate(dev, PCI_POWERSTATE_D0);
    }

    if (sdshc_timeout_interval == 0) {
	sdshc_timeout_interval = DEFAULT_TIMEOUT_INTERVAL;
    }
    if (sdshc_polling_interval == 0) {
	sdshc_polling_interval = DEFAULT_POLLING_INTERVAL;
    }
    if (sdshc_wait_after_insertion == 0) {
	sdshc_wait_after_insertion = DEFAULT_WAIT_AFTER_INSERTION;
    }

    if (sc->quirk_device != 0) {
#if !defined(USE_TI_DUMMY_DRIVER)
	if ((sc->quirk_device >> 16) == 0x104c && sdshc_disable_ti_alt_if) {
	    device_t quirk = pci_find_device(0x104c, sc->quirk_device & 0x0ffff);
	    uint32_t reg;

	    if (quirk != NULL) {
		MSG((dev, "disabling TI's alternate FlushMedia controller\n"));
		reg = pci_read_config(quirk, 0x4c, 1);
		reg |= 0x02;
		pci_write_config(quirk, 0x4c, reg, 1);
	    }
	}
#endif
    }

    if ((error = sdshc_init(sc)) != 0) {
	goto cleanup;
    }

    LOG(0x10, (dev, "Alloc Intr\n"));
    sc->irq_rid = 0;
    sc->irq = bus_alloc_resource_any(dev, SYS_RES_IRQ, &sc->irq_rid,
				     RF_ACTIVE | RF_SHAREABLE);
    if (sc->irq == NULL) {
	MSG((dev, "could not allocate interrupt resource\n"));
	error = ENOMEM;
	goto cleanup;
    }

    /*
     * Hook our interrupt after all initialization is complete.
     */
    LOG(0x10, (dev, "Setup Intr\n"));
    error = bus_setup_intr(dev, sc->irq, INTR_TYPE_NET|INTR_MPSAFE ,
#if (__FreeBSD_version >= 700031)
			   NULL,
#endif
			   sdshc_intr, sc, &sc->ih);
    if (error != 0) {
	MSG((dev, "could not set up interrupt\n"));
	goto cleanup;
    }

    sc->cdev = make_dev(&sdshc_cdevsw, device_get_unit(dev),
			UID_ROOT, GID_OPERATOR, 0664,
			"%s", device_get_nameunit(sc->dev));

    for (cnt = 0; cnt < sc->num_slot; cnt++) {
	device_add_child(dev, "sdshcbus", cnt);
    }

    bus_generic_attach(dev);

    return 0;

 cleanup:
    return error;
}


static int
sdshc_detach(device_t dev)
{
    struct sdshc_softc *sc = device_get_softc(dev);
    device_t *children;
    int nchildren;
    int cnt;

    /* detach & delete all children */
    bus_generic_detach(dev);
    if (!device_get_children(dev, &children, &nchildren)) {
	for (cnt = 0; cnt < nchildren; cnt++) {
	    device_delete_child(dev, children[cnt]);
	}
	free(children, M_TEMP);
    }

/*    sdshc_stop(sc);*/

    if (sc->cdev != NULL) {
	destroy_dev(sc->cdev);
	sc->cdev = NULL;
    }

    if (sc->irq != NULL) {
	bus_teardown_intr(dev, sc->irq, sc->ih);
	bus_release_resource(dev, SYS_RES_IRQ, sc->irq_rid, sc->irq);
    }
    if (sc->mem != NULL) {
	bus_release_resource(dev, SYS_RES_MEMORY, sc->mem_rid[0], sc->mem);
    }

    mtx_destroy(&sc->mtx);

    return 0;
}


static int
sdshc_shutdown(device_t dev)
{
    struct sdshc_softc *sc = device_get_softc(dev);

    sdshc_stop(sc);

    return 0;
}


static int
sdshc_suspend(device_t dev)
{
    struct sdshc_softc *sc = device_get_softc(dev);

    sdshc_stop(sc);

    return 0;
}


static int
sdshc_resume(device_t dev)
{
    struct sdshc_softc *sc = device_get_softc(dev);

    return sdshc_init(sc);
}


static void
sdshc_intr(void *arg)
{
    struct sdshc_softc *sc = arg;
    u_int32_t slot_intr_status;
    device_t *children;
    int nchildren, cnt, slot;

    SD_LOCK(sc);

    slot_intr_status = SDSHC_READ2(sc, SLOT_INTR_STATUS);
    LOG(0x10, (sc->dev, "Intr occured (0x%02x)\n", slot_intr_status));

    if (!device_get_children(sc->dev, &children, &nchildren)) {
	for (cnt = 0; cnt < nchildren; cnt++) {
	    if (children[cnt] != NULL) {
		slot = device_get_unit(children[cnt]) + sc->slot_base;
		if ((slot_intr_status & (1UL << slot)) != 0) {
		    SDBUS_INTR(children[cnt]);
		}
	    }
	}
	free(children, M_TEMP);
    }

    SD_UNLOCK(sc);
}


/**/


static int
sdshc_init(void *arg)
{
	struct sdshc_softc *sc = arg;
	int error;
	
	if ((error = sdshc_config(sc)) != 0) {
	    (void)sdshc_reset(sc);
	    goto cleanup;
	}

cleanup:
	return error;
}


static int
sdshc_config(struct sdshc_softc *sc)
{
    device_t dev = sc->dev;
    u_int32_t reg, base, num;
#if defined(SD_DEBUG)
    u_int32_t cap, tmp;
#endif
    int cnt;
    int error;

    if ((error = sdshc_reset(sc))) {
	return error;
    }

    LOG(0x08, (sc->dev, "Initializing device\n"));

    /* enable bus-mastering */
    pci_enable_busmaster(dev);

#if defined(SD_DEBUG)
    if (0x10 <= sdshc_debug_level) {
	reg = pci_read_config(dev, PCIR_CLASS, 1);
	MSG((dev, "Basic class: 0x%02x (%s)\n", reg,
	     reg == 0x08 ? "General Peripheral" : "unknown"));
	reg = pci_read_config(dev, PCIR_SUBCLASS, 1);
	MSG((dev, "Sub class: 0x%02x (%s)\n", reg,
	     reg == 0x05 ? "for SD Host Controller" : "unknown"));
	reg = pci_read_config(dev, PCIR_PROGIF, 1);
	MSG((dev, "Interface Code: 0x%02x (%s)\n", reg,
	     reg == 0x00 ? "Standard Host not supported DMA" :
	     reg == 0x01 ? "Standard Host supported DMA" :
	     reg == 0x02 ? "Vendor unique SD Host Controller" :
	     "unknown"));
	reg = pci_read_config(dev, PCIR_REVID, 1);
	MSG((dev, "Revision ID: 0x%02x\n", reg));
	reg = pci_read_config(dev, PCIR_HDRTYPE, 1);
	tmp = reg & PCIM_HDRTYPE;
	MSG((dev, "Header Type: 0x%02x (%s%s)\n", reg,
	     reg & PCIM_MFDEV ? "Multi Function, " : "",
	     tmp == PCIM_HDRTYPE_NORMAL ? "Normal device" :
	     tmp == PCIM_HDRTYPE_BRIDGE ? "Bridge" :
	     tmp == PCIM_HDRTYPE_CARDBUS ? "CardBus" :
	     "unknown"));

	reg = pci_read_config(dev, PCIR_SUBVEND_0, 2);
	MSG((dev, "Subsystem Vendor ID: 0x%04x\n", reg));
	reg = pci_read_config(dev, PCIR_SUBDEV_0, 2);
	MSG((dev, "Subsystem Device ID: 0x%04x\n", reg));
    }

    cap = pci_read_config(dev, PCIR_CAP_PTR, 1);
    LOG(0x10, (dev, "Capability Pointer: 0x%02x\n", cap));
    LOG(0x10, (dev, "Interrupt Line: %d\n",
	       pci_read_config(dev, PCIR_INTLINE, 1)));
    LOG(0x10, (dev, "Interrupt Pin: %d\n",
	       pci_read_config(dev, PCIR_INTPIN, 1)));
#endif

    reg = pci_read_config(dev, SDSHC_PCIR_SLOTINFO, 1);
    if (5 < (num = ((reg >> 4) & 0x07))) {
	MSG((dev, "Invalid number of slots (%d)\n", num));
	return ENXIO;
    }
    sc->num_slot = ++num;
    if (5 < (base = (reg & 0x07))) {
	MSG((dev,
	     "First Base Address Register Number (%d) is invalid\n", base));
	return ENXIO;
    }
    sc->slot_base = base;
    LOG(0x10, (dev, "Number Of Slots: %d\n", num));
    LOG(0x10, (dev, "First Base Address Register Number: %d\n", base));

    for (cnt = 0; cnt < num; cnt++) {
	sc->mem_rid[cnt] = PCIR_BAR(base + cnt);
#if defined(SD_DEBUG)
	LOG(0x10, (dev, "Base Address Register #%d: 0x%08x\n",
		   cnt, pci_read_config(dev, sc->mem_rid[cnt], 4)));
#endif
    }

    sc->mem_rid[0] = PCIR_BAR(base);
    sc->mem = bus_alloc_resource(dev, SYS_RES_MEMORY, &sc->mem_rid[0],
				 0ul, ~0ul, 256, RF_ACTIVE);
    if (sc->mem == NULL) {
	MSG((dev, "could not allocate memory resource\n"));
	return ENOMEM;
    }
    sc->bt = rman_get_bustag(sc->mem);
    sc->bh = rman_get_bushandle(sc->mem);

#if defined(SD_DEBUG)
    LOG(0x10, (dev, "Capability ID: 0x%02x\n",
	       pci_read_config(dev, cap + SDSHC_PCIR_CAP_CAPID, 1)));
    LOG(0x10, (dev, "Next Item Ptr: 0x%02x\n",
	       pci_read_config(dev, cap + SDSHC_PCIR_CAP_NEXTPTR, 1)));
    LOG(0x10, (dev, "Power Management Capabilities: 0x%04x\n",
	       pci_read_config(dev, cap + SDSHC_PCIR_CAP_PMC, 1)));
    LOG(0x10, (dev, "Power Management Control/Status: 0x%04x\n",
	       pci_read_config(dev, cap + SDSHC_PCIR_CAP_PMCSR, 1)));
    LOG(0x10, (dev, "PMCSR PCI to PCI Bridge Support: 0x%02x\n",
	       pci_read_config(dev, cap + SDSHC_PCIR_CAP_PMCSR_BSE, 1)));
    LOG(0x10, (dev, "Data: 0x%02x\n",
	       pci_read_config(dev, cap + SDSHC_PCIR_CAP_DATA, 1)));

    if (0x10 <= sdshc_debug_level) {
	reg = SDSHC_READ2(sc, HOST_CTLR_VER);
	MSG((dev, "Vendor Version Number: 0x%02x\n", (reg >> 8) & 0x0ff));
	tmp = reg & 0x0ff;
	MSG((dev,
	     "Specification Version Number: %s\n",
	     tmp == 0x00 ? "SD Host Specification Ver. 1.0" :
	     "(Reserved)"));
    }
#endif

    return 0;
}


static int
sdshc_reset(struct sdshc_softc *sc)
{
#if defined(SDSHC_CHECK_RESET_DONE)
    int cnt;
#endif

    LOG(0x08, (sc->dev, "Resetting device...\n"));
    SD_LOCK(sc);
    SDSHC_WRITE1(sc, RESET, SDSHC_HSR_RESET_ALL);
#if defined(SDSHC_CHECK_RESET_DONE)
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHC_READ1(sc, RESET) & SDSHC_HSR_RESET_MASK) == 0) {
	    LOG(0x01, (sc->dev, "Controller reset done\n"));
	    SD_UNLOCK(sc);
	    return 0;
	}
	msleep(sc, &sc->mtx, PPAUSE, "sdshc_reset", sdshc_polling_interval);
    }
    SD_UNLOCK(sc);
    LOG(0x01, (sc->dev, "Controller reset failure\n"));
    return ENXIO;
#else
    msleep(sc, &sc->mtx, PPAUSE, "sdshc_reset", hz / 20);
    SD_UNLOCK(sc);
    return 0;
#endif
}


static void
sdshc_stop(struct sdshc_softc *sc)
{
/*
	int slot_no;

	SD_LOCK(sc);
	for (slot_no = 0; slot_no < sc->num_slot; slot_no++) {
	    sdshc_disable_intr(sc, slot_no);
	    sdshc_card_powerdown(sc, slot_no);
	    sc->ssc[slot_no].card_inserted = SDDEV_INVALID;
	}
	SD_UNLOCK(sc);
*/
}


/* */

static int
sdshc_open(struct cdev *dev, int flag, int fmt, struct thread *td)
{
    int unit = SDSHC_UNIT(dev);
    struct sdshc_softc *sc = SDSHC_SOFTC(unit);

    if (sc == NULL) {
	return ENXIO;
    }

    /* Disallow multiple opens */
    if (sc->flag & SDSHC_STAT_OPENED) {
	return EBUSY;
    }

    device_busy(devclass_get_device(sdshc_devclass, unit));

    sc->flag |= SDSHC_STAT_OPENED;

    return 0;
}


static int
sdshc_close(struct cdev *dev, int flag, int fmt, struct thread *td)
{
    int unit = SDSHC_UNIT(dev);
    struct sdshc_softc *sc = SDSHC_SOFTC(unit);

    device_unbusy(devclass_get_device(sdshc_devclass, unit));
    sc->flag &= ~SDSHC_STAT_OPENED;

    return 0;
}


/*
static int
sdshc_poll(struct cdev *dev, int events, struct thread *td)
{
}


static void
sdshc_purge(struct cdev *dev)
{
}
*/


MODULE_DEPEND(sdshc, pci,  1, 1, 1);
